import numpy as np
from lmfit import Parameters, minimize


def _heartrate_model_predict(model_params, power):
    power = power + power.cumsum() * model_params['hr_drift']
    hr_lin = power * model_params['dhr'] + model_params['hr_rest']
    hr_ss = np.minimum(model_params['hr_max'], hr_lin)

    dhr = np.insert(np.diff(power), 0, 0)
    tau = np.where((dhr <= 0), model_params['tau_fall'], model_params['tau_rise'])

    predicted_hr = []
    hr_previous = model_params['hr_rest']
    for h, t in zip(hr_ss, tau):
        hr_previous = hr_previous + (h - hr_previous)/t
        predicted_hr.append(hr_previous)

    return predicted_hr


def _heartrate_model_residuals(model_params, power, heartrate):
    model = _heartrate_model_predict(model_params, power)
    return heartrate - model


def heartrate_model(heartrate, power, **kwargs):
    """
    Source:
    De Smet et al., Heart rate modelling as a potential physical finess assessment for runners and cyclists.
    http://ceur-ws.org/Vol-1842/paper_13.pdf
    """
    # Initial model parameters
    model_params = Parameters()
    model_params.add_many(
        ('hr_rest', kwargs.get('hr_rest', 75)),
        ('hr_max', kwargs.get('hr_max', 200)),
        ('dhr', kwargs.get('dhr', 0.30)),
        ('tau_rise', kwargs.get('tau_rise', 24)),
        ('tau_fall', kwargs.get('tau_fall', 30)),
        ('hr_drift', kwargs.get('hr_drift', 3 * 10**-5)),
    )

    model = minimize(
        fcn=_heartrate_model_residuals,
        params=model_params,
        method='nelder',  # Nelder-Mead
        args=(power, heartrate)
    )

    predictions = _heartrate_model_predict(model.params, power)

    return model, predictions



from .algorithms import heartrate_models, main, w_prime_balance
from .base import BaseWorkoutDataFrame
from .helpers import requires


class WorkoutDataFrame(BaseWorkoutDataFrame):
    _metadata = ['athlete']

    @requires(columns=['power'])
    def compute_mean_max_power(self):
        return main.mean_max_power(self.power)

    @requires(columns=['power'])
    def compute_weighted_average_power(self):
        return main.weighted_average_power(self.power)

    @requires(columns=['power'], athlete=['weight'])
    def compute_power_per_kg(self):
        return main.power_per_kg(self.power, self.athlete.weight)

    @requires(columns=['power'], athlete=['cp', 'w_prime'])
    def compute_w_prime_balance(self, algorithm=None, *args, **kwargs):
        return w_prime_balance.w_prime_balance(self.power, self.athlete.cp,
            self.athlete.w_prime, algorithm, *args, **kwargs)

    @requires(columns=['power'])
    def compute_mean_max_bests(self, duration, amount):
        return main.mean_max_bests(self.power, duration, amount)

    @requires(columns=['power', 'heartrate'])
    def compute_heartrate_model(self):
        return heartrate_models.heartrate_model(self.heartrate, self.power)


class Athlete:
    def __init__(self, name=None, sex=None, weight=None, dob=None, ftp=None,
            cp=None, w_prime=None):
        self.name = name
        self.sex = sex
        self.weight = weight
        self.dob = dob
        self.ftp = ftp
        self.cp = cp
        self.w_prime = w_prime



class MissingDataException(Exception):
        def __init__(self, func, missing_fields):
            self.message = 'Method \'{}\' is missing data: {}'.format(
                    func.__name__,
                    ', '.join(missing_fields))

            super(MissingDataException, self).__init__(self.message)


class WorkoutDataFrameValidationException(Exception):
        def __init__(self, errors):
            self.message = 'Validation failed at:\n' + '\n'.join(errors)
            super(WorkoutDataFrameValidationException, self).__init__(self.message)



from .exceptions import MissingDataException


def requires(columns=None, athlete=None):
    def requires_wrapper(func):
        def requires_decorator(*args, **kwargs):
            wdf = args[0]
            missing_data = set()

            if columns:
                missing_data.update(set(columns) - set(list(wdf)))

            if athlete:
                missing_data.update(
                    {i for i in athlete if getattr(wdf.athlete, i, None) is None})

            if missing_data:
                raise MissingDataException(func, missing_data)

            return func(*args, **kwargs)

        return requires_decorator

    return requires_wrapper



import numpy as np

from .exceptions import WorkoutDataFrameValidationException

VALIDATION_COLUMNS = ['time', 'power', 'heartrate', 'cadence', 'speed']

COLUMN_DTYPES = {
        'time': np.int,
        'power': np.int,
        'heartrate': np.int,
        'cadence': np.int,
        'speed': np.int}

COLUMN_MIN_VALUE = {
        'time': 0,
        'power': 0,
        'heartrate': 0,
        'cadence': 0,
        'speed': 0}

COLUMN_MAX_VALUE = {
        'time': np.inf,
        'power': 3000,
        'heartrate': 300,
        'cadence': 300,
        'speed': 0}


class ValidationMixin(object):
    def is_valid(self):
        errors = list()

        errors += self._validate_index()
        errors += self._validate_columns()

        if errors:
            raise WorkoutDataFrameValidationException(errors)
        return True
    
    def _validate_columns(self):
        errors = list()
        columns = set(VALIDATION_COLUMNS) & set(list(self))

        for c in columns:
            errors += self._validate_series(
                series=self[c],
                dtype=COLUMN_DTYPES[c], 
                min_value=COLUMN_MIN_VALUE[c],
                max_value=COLUMN_MAX_VALUE[c])

        return errors

    @classmethod
    def _validate_series(cls, series, dtype, min_value, max_value):
        errors = list()
        if series.dtype != dtype:
            errors.append(
                'Column \'{}\' is not of dtype \'{}\''.format(series.name, dtype))
        if series.min() < min_value:
            errors.append('Column \'{}\' has values < {}'.format(series.name, min_value))
        if series.max() > max_value:
            errors.append('Column \'{}\' has values > {}'.format(series.name, max_value))
        return errors

    def _validate_index(self):
        errors = list()

        if not self.index.name == 'time':
            errors.append(
                'Index should be set to \'time\'. See \'pandas.DataFrame.set_index\'')
            return errors

        # @TODO change diff to calculation that's meaningful for time object
        # @TODO also change time dtype and min max
        index_diff = self.index.to_series().diff(1)
        if index_diff.where(index_diff != 1).where(index_diff != np.nan).count():
            errors.append('Sample rate is not (consistent) 1Hz')

        return errors



import math
from collections import namedtuple

from lmfit import minimize, Parameters
import numpy as np
import pandas as pd

from . import critical_power
from . import heartrate_models
from . import w_prime_balance


def mean_max_power(power):
    mmp = []

    for i in range(len(power)):
        mmp.append(power.rolling(i+1).mean().max())

    return pd.Series(mmp)


DataPoint = namedtuple('DataPoint', ['index', 'value'])


def mean_max_bests(power, duration, amount):
    moving_average = power.rolling(duration).mean()
    length = len(moving_average)
    mean_max_bests = []

    for i in range(amount):
        if moving_average.isnull().all():
            mean_max_bests.append(DataPoint(np.nan, np.nan))
            continue

        max_value = moving_average.max()
        max_index = moving_average.idxmax()
        mean_max_bests.append(DataPoint(max_index, max_value))

        # Set moving averages that overlap with last found max to np.nan
        overlap_min_index = max(0, max_index-duration)
        overlap_max_index = min(length, max_index+duration)
        moving_average.loc[overlap_min_index:overlap_max_index] = np.nan

    return pd.Series(mean_max_bests)


def weighted_average_power(power):
    wap = power.rolling(30).mean().pow(4).mean().__pow__(1/4)
    return wap


def power_per_kg(power, weight):
    ppkg = power / weight
    return ppkg



import pandas as pd

from .mixins import ValidationMixin


class BaseWorkoutDataFrame(pd.DataFrame, ValidationMixin):
    _metadata = ['athlete']

    @property
    def _constructor(self):
        return self.__class__ # A less hacky way would be appreciated...



from collections import OrderedDict

import numpy as np
from scipy.optimize import curve_fit


def two_parameter_non_linear_predict(t, cp, w_prime):
    return cp + w_prime / t


def three_parameter_non_linear_predict(t, cp, w_prime, p_max):
    return w_prime / (t + (w_prime / (p_max - cp))) + cp


def extended_5_3_predict(t, power_anaerobic_alactic, power_anaerobic_decay,
                     cp, tau_delay, cp_delay, cp_decay, cp_decay_delay, tau):
    """
    Credits to Damien Grauser. Source:
    https://github.com/GoldenCheetah/GoldenCheetah/blob/master/src/Metrics/ExtendedCriticalPower.cpp
    """
    model = (
        (
            (
                power_anaerobic_alactic
            ) * (
                1.20 - 0.20 * np.exp(-1 * t)
            ) * (
                np.exp(power_anaerobic_decay * t)
            )
        ) + (
            (
                cp
            ) * (
                1 - np.exp(cp_delay * t)
            ) * (
                1 + cp_decay * np.exp(cp_decay_delay / t)
            ) * (
                1 - np.exp(tau_delay * t)
            ) * (
                1 + tau / t
            )
        )
    )

    model = np.nan_to_num(model)
    model[model > 2000] = 0
    model[model < 1] = 0

    return model


def extended_7_3_predict(t, power_anaerobic_alactic, power_anaerobic_decay,
                     cp, tau_delay, cp_delay, cp_decay, cp_decay_delay, tau):
    """
    Credits to Damien Grauser. Source:
    https://github.com/GoldenCheetah/GoldenCheetah/blob/master/src/Metrics/ExtendedCriticalPower.cpp
    """
    model = (
        (
            (
                power_anaerobic_alactic
            ) * (
                (
                    np.exp(power_anaerobic_decay)
                ) * (
                    np.power(t, power_anaerobic_alactic)
                )
            )
        ) + (
            (
                cp * (1 - np.exp(tau_delay * t))
            ) * (
                1 - np.exp(cp_delay * t)
            ) * (
                1 + cp_decay * np.exp(cp_decay_delay / t)
            ) * (
                1 + tau / t
            )
        )
    )

    model = np.nan_to_num(model)
    model[model > 2000] = 0
    model[model < 1] = 0

    return model


def model_fit(x, y, model='extended_5_3'):
    if model == '2_parameter_non_linear':
        predict_func = two_parameter_non_linear_predict
        initial_model_params = OrderedDict(
            cp =300,
            w_prime=20000
        )
    if model == '3_parameter_non_linear':
        predict_func = three_parameter_non_linear_predict
        initial_model_params = OrderedDict(
            cp =300,
            w_prime=20000,
            p_max=1000
        )
    elif model == 'extended_5_3':
        predict_func = extended_5_3_predict
        initial_model_params = OrderedDict(
            power_anaerobic_alactic=811,
            power_anaerobic_decay=-2,
            cp=280,
            cp_delay=-0.9,
            cp_decay=-0.583,
            cp_decay_delay=-180,
            tau=1.208,
            tau_delay=-4.8
        )
    elif model == 'extended_7_3':
        predict_func = extended_7_3_predict
        initial_model_params = OrderedDict(
            power_anaerobic_alactic=811,
            power_anaerobic_decay=-2,
            cp=280,
            cp_delay=-0.9,
            cp_decay=-0.583,
            cp_decay_delay=-180,
            tau=1.208,
            tau_delay=-4.8
        )

    initial_param_array = np.array(list(initial_model_params.values()))

    model_params, _ = curve_fit(
        f=predict_func,
        xdata=x,
        ydata=y,
        p0=initial_param_array,
    )

    fitted_model_params = dict(zip(initial_model_params.keys(), model_params))

    return fitted_model_params



from setuptools import setup, find_packages

setup(
    name = 'sweat',
    packages = find_packages(),
    version = '0.1.0',
    description = 'Workout analysis',
    author='Aart Goossens',
    author_email='aart@goossens.me',
    url='https://github.com/GoldenCheetah/sweatpy',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
    ]
)



import unittest

import numpy as np
import pandas as pd
import pytest

from sweat import exceptions, models


@pytest.fixture
def wdf():
    data = {
        'time': range(10),
        'heartrate': range(10),
        'power': range(10)}
    athlete = models.Athlete(name='Chris', weight=80, ftp=300)
    wdf = models.WorkoutDataFrame(data)
    wdf = wdf.set_index('time')
    wdf.athlete = athlete
    return wdf


@pytest.fixture
def wdf_small():
    athlete = models.Athlete(cp=200, w_prime=20000)
    wdf = models.WorkoutDataFrame(
        pd.read_csv('tests/example_files/workout_1_short.csv')
    )
    wdf = wdf.set_index('time')
    wdf.athlete = athlete
    return wdf


@pytest.fixture
def wdf_big():
    athlete = models.Athlete(cp=200, w_prime=20000, weight=80)
    wdf = models.WorkoutDataFrame(
        pd.read_csv('tests/example_files/workout_1.csv')
    )
    wdf = wdf.set_index('time')
    wdf.athlete = athlete
    return wdf


class TestAthlete:
    def setup(self):
        self.athlete = models.Athlete(name='Chris', ftp=300)

    def test_empty_init(self):
        athlete = models.Athlete()
        assert hasattr(athlete, 'name')
        assert hasattr(athlete, 'sex')
        assert hasattr(athlete, 'weight')
        assert hasattr(athlete, 'dob')
        assert hasattr(athlete, 'ftp')
        assert hasattr(athlete, 'cp')
        assert hasattr(athlete, 'w_prime')

    def test_init(self):
        assert self.athlete.name == 'Chris'
        assert self.athlete.ftp == 300


class TestWorkoutDataFrame:
    def test_empty_init(self):
        wdf = models.WorkoutDataFrame()

        assert isinstance(wdf, pd.DataFrame)
        assert isinstance(wdf, models.WorkoutDataFrame)

    def test_init(self, wdf):
        assert 'power' in list(wdf)
        assert 'heartrate' in list(wdf)
        assert len(wdf) == 10
        assert hasattr(wdf, 'athlete')

    def test_slicing(self, wdf):
        new_wdf = wdf[1:5]

        assert isinstance(new_wdf, models.WorkoutDataFrame)

    def test_metadata_propagation(self, wdf):
        assert wdf[1:5].athlete.name == 'Chris'
        assert wdf.iloc[[0, 1], :].athlete.name == 'Chris'
        assert wdf[['power']].athlete.name == 'Chris'

    def test_is_valid(self, wdf):
        assert wdf.is_valid()

    def test_is_valid_missing_time_index(self, wdf):
        wdf.index.name = 'not_time'

        with pytest.raises(exceptions.WorkoutDataFrameValidationException):
            wdf.is_valid()

    def test_is_valid_invalid_sample_rate(self):
        data = {
            'time': range(0, 20, 2),
            'heartrate': range(10),
            'power': range(10)}
        wdf = models.WorkoutDataFrame(data)
        wdf = wdf.set_index('time')

        with pytest.raises(exceptions.WorkoutDataFrameValidationException) as e:
            wdf.is_valid()
            assert e.message == '[.\n]*Sample rate is not \(consistent\) 1Hz[.\n]*'

    def test_is_valid_invalid_dtype(self):
        data = {
            'time': range(10),
            'heartrate': np.arange(0, 15, 1.5),
            'power': range(10)}
        wdf = models.WorkoutDataFrame(data)
        wdf = wdf.set_index('time')

        with pytest.raises(exceptions.WorkoutDataFrameValidationException) as e:
            wdf.is_valid()
            assert e.message == '[.\n]*Column \'heartrate\' is not of dtype[.\n]*'

    def test_is_valid_invalid_min_value(self):
        data = {
            'time': range(10),
            'heartrate': range(-10, 0),
            'power': range(10)}
        wdf = models.WorkoutDataFrame(data)
        wdf = wdf.set_index('time')

        with pytest.raises(exceptions.WorkoutDataFrameValidationException) as e:
            wdf.is_valid()
            assert e.message == '[.\n]*Column \'heartrate\' has values < 0[.\n]*'

    def test_is_valid_invalid_max_value(self):
        data = {
            'time': range(10),
            'heartrate': range(10),
            'power': range(10000, 10010)}
        wdf = models.WorkoutDataFrame(data)
        wdf = wdf.set_index('time')

        with pytest.raises(exceptions.WorkoutDataFrameValidationException) as e:
            wdf.is_valid()
            assert e.message == '[.\n]*Column \'power\' has values > 3000[.\n]*'

    def test_compute_mean_max_power(self, wdf_big):
        mmp = wdf_big.compute_mean_max_power()

        assert mmp.iloc[1] == 263.0
        assert mmp.iloc[300] == 209.37209302325581

    def test_compute_mean_max_power_missing_power(self, wdf):
        del wdf['power']

        with pytest.raises(exceptions.MissingDataException):
            assert wdf.compute_mean_max_power() is None

    def test_compute_weighted_average_power(self, wdf_big):
        assert wdf_big.compute_weighted_average_power() == 156.24624656343036

    def test_compute_weighted_average_power_missing_power(self, wdf_big):
        del wdf_big['power']

        with pytest.raises(exceptions.MissingDataException):
            wdf_big.compute_weighted_average_power()

    def test_compute_power_per_kg(self, wdf_big):
        # wdf_big.athlete.ftp = 300
        ppkg = wdf_big.compute_power_per_kg()

        assert ppkg[1] == 1.1625000000000001
        assert ppkg[100] == 1.0125

    def test_compute_power_per_kg_missing_weight(self, wdf):
        wdf.athlete.weight = None

        with pytest.raises(exceptions.MissingDataException):
            wdf.compute_power_per_kg()

    @pytest.mark.parametrize("test_input,expected", [
        (dict(), 19006.732787629098),
        (dict(algorithm='waterworth'), 19006.732787629098),
        (dict(algorithm='waterworth', tau_value=500), 19118.509110305589),
        (dict(algorithm='waterworth', tau_dynamic=True), 18989.003862950944),
        (dict(algorithm='skiba'), 19006.732787629102),
        (dict(algorithm='skiba', tau_value=500), 19118.509110305589),
        (dict(algorithm='skiba', tau_dynamic=True), 19023.724199486114),
        (dict(algorithm='froncioni-skiba-clarke'), 19170.066066056541),
    ])
    def test_compute_w_prime_balance(self, wdf_small, test_input, expected):
        w_balance = wdf_small.compute_w_prime_balance(**test_input)

        assert w_balance.iloc[800] == expected

    def test_compute_mean_max_bests(self, wdf_big):
        result = wdf_big.compute_mean_max_bests(60, 3)

        assert len(result) == 3
        assert result[0] == (2038, 215.13333333333333)
        assert result[1] == (2236, 210.48333333333332)
        assert result[2] == (2159, 208.93333333333334)

    def test_compute_mean_max_bests_only_one_result(self, wdf_big):
        result = wdf_big.compute_mean_max_bests(3000, 2)

        assert len(result) == 2
        assert not result[0] == (np.nan, np.nan)
        assert result[1] == (np.nan, np.nan)

    def test_compute_mean_max_bests_no_results(self, wdf_big):
        result = wdf_big.compute_mean_max_bests(10000, 1)

        assert len(result) == 1
        assert result[0] == (np.nan, np.nan)

    def test_compute_heartrate_model(self, wdf_big):
        model, predictions = wdf_big.compute_heartrate_model()

        assert len(predictions) == len(wdf_big)
        assert model.params['hr_rest'].value == 86.552894067943839
        assert model.params['hr_max'].value == 223.29248085198429
        assert model.params['dhr'].value == 0.26860492517685708
        assert model.params['tau_rise'].value == 51.806209961533796
        assert model.params['tau_fall'].value == 155.18880949951827
        assert model.params['hr_drift'].value == 7.9899370717145379 * 10**-5
